/*
    Modbus.cpp - Source for Modbus Base Library
    Copyright (C) 2014 André Sarmento Barbosa
    Copyright (C) 2023 Pascal JEAN aka epsilonrt
*/
#include "Modbus.h"

//-------------------------------------------------------------------------------
Modbus::Modbus() {
  _regs_head = 0;
  _regs_last = 0;
  _additional_data = 0;
}

//-------------------------------------------------------------------------------
// private
TRegister *Modbus::searchRegister (word address) {
  TRegister *reg = _regs_head;
  //if there is no register configured, bail
  if (reg == 0) {
    return (0);
  }
  //scan through the linked list until the end of the list or the register is found.
  //return the pointer.
  do {
    if (reg->address == address) {
      return (reg);
    }
    reg = reg->next;
  }
  while (reg);
  return (0);
}

//-------------------------------------------------------------------------------
// private
void Modbus::addReg (word address, word value) {
  TRegister *newreg;

  newreg = (TRegister *) malloc (sizeof (TRegister));
  newreg->address = address;
  newreg->value   = value;
  newreg->edata   = 0;
  newreg->next    = 0;

  if (_regs_head == 0) {
    _regs_head = newreg;
    _regs_last = _regs_head;
  }
  else {
    //Assign the last register's next pointer to newreg.
    _regs_last->next = newreg;
    //then make temp the last register in the list.
    _regs_last = newreg;
  }
}

//-------------------------------------------------------------------------------
// private
bool Modbus::setReg (word address, word value) {
  TRegister *reg;
  //search for the register address
  reg = searchRegister (address);
  //if found then assign the register value to the new value.
  if (reg) {
    reg->value = value;
    return true;
  }
  else {
    return false;
  }
}

//-------------------------------------------------------------------------------
// private
word Modbus::reg (word address) {
  TRegister *reg;
  reg = searchRegister (address);
  if (reg) {
    return (reg->value);
  }
  else {
    return (0);
  }
}

//-------------------------------------------------------------------------------
// private
bool  Modbus::setRegBounds (word address, word min, word max) {
  TRegister *reg;

  reg = searchRegister (address);

  if (reg) {

    if (!reg->edata) {

      reg->edata = new TExtData;
    }
    if (reg->edata) {

      reg->edata->min = min;
      reg->edata->max = max;
      return true;
    }
  }
  return false;
}

//-------------------------------------------------------------------------------
// private
bool  Modbus::regOutOfBounds (word address, word value) {
  TRegister *reg;

  reg = searchRegister (address);

  if (reg) {

    if (reg->edata) {

      return (value < reg->edata->min) || (value > reg->edata->max);
    }
  }
  return false;
}

//-------------------------------------------------------------------------------
// protected
void Modbus::receivePDU (byte *frame) {
  byte fcode  = frame[0];
  word field1 = (word) frame[1] << 8 | (word) frame[2];
  word field2 = (word) frame[3] << 8 | (word) frame[4];

  switch (fcode) {

    case MB_FC_WRITE_REG:
      //field1 = reg, field2 = value
      writeSingleRegister (field1, field2);
      break;

    case MB_FC_READ_REGS:
      //field1 = startreg, field2 = numregs
      readRegisters (field1, field2);
      break;

    case MB_FC_WRITE_REGS:
      //field1 = startreg, field2 = status
      writeMultipleRegisters (frame, field1, field2, frame[5]);
      break;

      #ifndef USE_HOLDING_REGISTERS_ONLY

    case MB_FC_REPORT_SERVER_ID:
      reportServerId();
      break;

    case MB_FC_READ_COILS:
      //field1 = startreg, field2 = numregs
      readCoils (field1, field2);
      break;

    case MB_FC_READ_INPUT_STAT:
      //field1 = startreg, field2 = numregs
      readInputStatus (field1, field2);
      break;

    case MB_FC_READ_INPUT_REGS:
      //field1 = startreg, field2 = numregs
      readInputRegisters (field1, field2);
      break;

    case MB_FC_WRITE_COIL:
      //field1 = reg, field2 = status
      writeSingleCoil (field1, field2);
      break;

    case MB_FC_WRITE_COILS:
      //field1 = startreg, field2 = numoutputs
      writeMultipleCoils (frame, field1, field2, frame[5]);
      break;

      #endif

    default:
      exceptionResponse (fcode, MB_EX_ILLEGAL_FUNCTION);
  }
}

//-------------------------------------------------------------------------------
// private
void Modbus::exceptionResponse (byte fcode, byte excode) {
  //Clean frame buffer
  free (_frame);
  _len = 2;
  _frame = (byte *) malloc (_len);
  _frame[0] = fcode + 0x80;
  _frame[1] = excode;

  _reply = MB_REPLY_NORMAL;
}

/* Func 06: Write Single Register
    Request
      Function code     1 Byte  0x06
      Register Address  2 Bytes 0x0000 to 0xFFFF
      Register Value    2 Bytes 0x0000 to 0xFFFF
    Response
      Function code     1 Byte  0x06
      Register Address  2 Bytes 0x0000 to 0xFFFF
      Register Value    2 Bytes 0x0000 to 0xFFFF
*/
//-------------------------------------------------------------------------------
// private
void Modbus::writeSingleRegister (word reg, word value) {

  if (hregOutOfBounds (reg, value)) {
    exceptionResponse (MB_FC_WRITE_REG, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address and execute (reg exists?)
  if (!setHreg (reg, value)) {
    exceptionResponse (MB_FC_WRITE_REG, MB_EX_ILLEGAL_ADDRESS);
    return;
  }

  //Check for failure ? Why ???
  /*
    if (hreg (reg) != value) {
    exceptionResponse (MB_FC_WRITE_REG, MB_EX_SLAVE_FAILURE);
    return;
    }
  */
  _reply = MB_REPLY_ECHO; // reply with received frame
}

/*
  Func 03: Read Holding Registers
    Request
      Function code         1 Byte  0x03
      Starting Address      2 Bytes 0x0000 to 0xFFFF
      Quantity of Registers 2 Bytes 1 to 125 (0x7D)
    Response
      Function code         1 Byte  0x03
      Byte count            1 Byte  2 x N*
      Register value        N* x 2 Bytes
     N = Quantity of Registers
*/
//-------------------------------------------------------------------------------
// private
void Modbus::readRegisters (word startreg, word numregs) {
  //Check value (numregs)
  if (numregs < 1 || numregs > 125) {
    exceptionResponse (MB_FC_READ_REGS, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address (startreg...startreg + numregs - 1)
  for (word k = 0; k < numregs; k++) {
    if (!searchRegister (startreg + TRegister::HregOffset + k)) {
      exceptionResponse (MB_FC_WRITE_REGS, MB_EX_ILLEGAL_ADDRESS);
      return;
    }
  }

  //Clean frame buffer
  free (_frame);
  _len = 0;

  //calculate the query reply message length
  //for each register queried add 2 bytes
  _len = 2 + numregs * 2;

  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_READ_REGS, MB_EX_SLAVE_FAILURE);
    return;
  }

  _frame[0] = MB_FC_READ_REGS;
  _frame[1] = _len - 2;   //byte count

  word val;
  word i = 0;
  while (numregs--) {
    //retrieve the value from the register bank for the current register
    val = hreg (startreg + i);
    //write the high byte of the register value
    _frame[2 + i * 2]  = val >> 8;
    //write the low byte of the register value
    _frame[3 + i * 2] = val & 0xFF;
    i++;
  }

  _reply = MB_REPLY_NORMAL;
}

/* Func 16: Write Multiple registers
    Request
      Function code         1 Byte        0x10
      Starting Address      2 Bytes       0x0000 to 0xFFFF
      Quantity of Registers 2 Bytes       0x0001 to 0x007B
      Byte Count            1 Byte        2 x N*
      Registers Value       N* x 2 Bytes  value
       N = Quantity of Registers
    Response
      Function code         1 Byte        0x10
      Starting Address      2 Bytes       0x0000 to 0xFFFF
      Quantity of Registers 2 Bytes       1 to 123 (0x7B)
*/
//-------------------------------------------------------------------------------
// private
void Modbus::writeMultipleRegisters (byte *frame, word startreg, word numoutputs, byte bytecount) {
  //Check value
  if (numoutputs < 1 || numoutputs > 123 || bytecount != 2 * numoutputs) {
    exceptionResponse (MB_FC_WRITE_REGS, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address (startreg...startreg + numregs - 1)
  for (word k = 0; k < numoutputs; k++) {
    if (!searchRegister (startreg + TRegister::HregOffset + k)) {
      exceptionResponse (MB_FC_WRITE_REGS, MB_EX_ILLEGAL_ADDRESS);
      return;
    }
  }

  //Clean frame buffer
  free (_frame);
  _len = 5;
  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_WRITE_REGS, MB_EX_SLAVE_FAILURE);
    return;
  }

  _frame[0] = MB_FC_WRITE_REGS;
  _frame[1] = startreg >> 8;
  _frame[2] = startreg & 0x00FF;
  _frame[3] = numoutputs >> 8;
  _frame[4] = numoutputs & 0x00FF;

  word val;
  word i = 0;
  while (numoutputs--) {
    val = (word) frame[6 + i * 2] << 8 | (word) frame[7 + i * 2];
    if (hregOutOfBounds (startreg + i, val)) {
      exceptionResponse (MB_FC_WRITE_REG, MB_EX_ILLEGAL_VALUE);
      return;
    }
    setHreg (startreg + i, val);
    i++;
  }

  _reply = MB_REPLY_NORMAL;
}

#ifndef USE_HOLDING_REGISTERS_ONLY
//-------------------------------------------------------------------------------

/* Func 01:  Read Coils
    Request
      Function code     1 Byte  0x01
      Starting Address  2 Bytes 0x0000 to 0xFFFF
      Quantity of coils 2 Bytes 1 to 2000 (0x7D0)
    Response
      Function code     1 Byte  0x01
      Byte count        1 Byte  N*
      Coil Status       n Byte  n = N or N+1
     N = Quantity of Outputs / 8, if the remainder is different of 0 -> N = N+1
*/
//-------------------------------------------------------------------------------
// private
void Modbus::readCoils (word startreg, word numregs) {
  //Check value (numregs)
  if (numregs < 1 || numregs > 2000) {
    exceptionResponse (MB_FC_READ_COILS, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //When I check all registers in range I got errors in ScadaBR
  //I think that ScadaBR request more than one in the single request
  //when you have more then one datapoint configured from same type.
  // epsilonrt -> We absolutely must not return values that do not exist !!

  //Check Address (startreg...startreg + numregs - 1)
  for (word k = 0; k < numregs; k++) {
    if (!searchRegister (startreg + TRegister::CoilOffset + k)) {
      exceptionResponse (MB_FC_WRITE_REGS, MB_EX_ILLEGAL_ADDRESS);
      return;
    }
  }

  //Clean frame buffer
  free (_frame);
  _len = 0;

  //Determine the message length = function type, byte count and
  //for each group of 8 registers the message length increases by 1
  _len = 2 + numregs / 8;
  if (numregs % 8) {
    _len++;  //Add 1 to the message length for the partial byte.
  }

  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_READ_COILS, MB_EX_SLAVE_FAILURE);
    return;
  }

  _frame[0] = MB_FC_READ_COILS;
  _frame[1] = _len - 2; //byte count (_len - function code and byte count)

  byte bitn = 0;
  word totregs = numregs;
  word i;
  while (numregs--) {
    i = (totregs - numregs) / 8;
    if (coil (startreg)) {
      bitSet (_frame[2 + i], bitn);
    }
    else {
      bitClear (_frame[2 + i], bitn);
    }
    //increment the bit index
    bitn++;
    if (bitn == 8) {
      bitn = 0;
    }
    //increment the register
    startreg++;
  }

  _reply = MB_REPLY_NORMAL;
}

/* Func 02: Read Discrete Inputs
    Request
      Function code       1 Byte      0x02
      Starting Address    2 Bytes     0x0000 to 0xFFFF
      Quantity of Inputs  2 Bytes     1 to 2000 (0x7D0)
    Response
      Function code       1 Byte      0x02
      Byte count          1 Byte      N*
      Input Status        N* x 1 Byte
     N = Quantity of Inputs / 8 if the remainder is different of 0  N = N+1
  Error
*/
//-------------------------------------------------------------------------------
// private
void Modbus::readInputStatus (word startreg, word numregs) {
  //Check value (numregs)
  if (numregs < 0x0001 || numregs > 0x07D0) {
    exceptionResponse (MB_FC_READ_INPUT_STAT, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address (startreg...startreg + numregs - 1)
  for (word k = 0; k < numregs; k++) {
    if (!searchRegister (startreg + TRegister::IstsOffset + k)) {
      exceptionResponse (MB_FC_WRITE_REGS, MB_EX_ILLEGAL_ADDRESS);
      return;
    }
  }

  //Clean frame buffer
  free (_frame);
  _len = 0;

  //Determine the message length = function type, byte count and
  //for each group of 8 registers the message length increases by 1
  _len = 2 + numregs / 8;
  if (numregs % 8) {
    _len++;  //Add 1 to the message length for the partial byte.
  }

  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_READ_INPUT_STAT, MB_EX_SLAVE_FAILURE);
    return;
  }

  _frame[0] = MB_FC_READ_INPUT_STAT;
  _frame[1] = _len - 2;

  byte bitn = 0;
  word totregs = numregs;
  word i;
  while (numregs--) {
    i = (totregs - numregs) / 8;
    if (ists (startreg)) {
      bitSet (_frame[2 + i], bitn);
    }
    else {
      bitClear (_frame[2 + i], bitn);
    }
    //increment the bit index
    bitn++;
    if (bitn == 8) {
      bitn = 0;
    }
    //increment the register
    startreg++;
  }

  _reply = MB_REPLY_NORMAL;
}

/* Func 04: Read Input Registers
    Request
      Function code               1 Byte        0x04
      Starting Address            2 Bytes       0x0000 to 0xFFFF
      Quantity of Input Registers 2 Bytes       0x0001 to 0x007D
    Response
      Function code               1 Byte        0x04
      Byte count                  1 Byte        2 x N*
      Input Registers             N* x 2 Bytes
     N = Quantity of Input Registers
*/
//-------------------------------------------------------------------------------
// private
void Modbus::readInputRegisters (word startreg, word numregs) {
  //Check value (numregs)
  if (numregs < 0x0001 || numregs > 0x007D) {
    exceptionResponse (MB_FC_READ_INPUT_REGS, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address (startreg...startreg + numregs - 1)
  for (word k = 0; k < numregs; k++) {
    if (!searchRegister (startreg + TRegister::IregOffset + k)) {
      exceptionResponse (MB_FC_WRITE_REGS, MB_EX_ILLEGAL_ADDRESS);
      return;
    }
  }

  //Clean frame buffer
  free (_frame);
  _len = 0;

  //calculate the query reply message length
  //for each register queried add 2 bytes
  _len = 2 + numregs * 2;

  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_READ_INPUT_REGS, MB_EX_SLAVE_FAILURE);
    return;
  }

  _frame[0] = MB_FC_READ_INPUT_REGS;
  _frame[1] = _len - 2;

  word val;
  word i = 0;
  while (numregs--) {
    //retrieve the value from the register bank for the current register
    val = ireg (startreg + i);
    //write the high byte of the register value
    _frame[2 + i * 2]  = val >> 8;
    //write the low byte of the register value
    _frame[3 + i * 2] = val & 0xFF;
    i++;
  }

  _reply = MB_REPLY_NORMAL;
}

/* Func 05: Write Single Coil
    Request
      Function code   1 Byte  0x05
      Output Address  2 Bytes 0x0000 to 0xFFFF
      Output Value    2 Bytes 0x0000 or 0xFF00
  Response
      Function code   1 Byte  0x05
      Output Address  2 Bytes 0x0000 to 0xFFFF
      Output Value    2 Bytes 0x0000 or 0xFF00
*/
//-------------------------------------------------------------------------------
// private
void Modbus::writeSingleCoil (word reg, word status) {
  //Check value (status)
  if (status != 0xFF00 && status != 0x0000) {
    exceptionResponse (MB_FC_WRITE_COIL, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address and execute (reg exists?)
  if (!setCoil (reg, status != 0)) {
    exceptionResponse (MB_FC_WRITE_COIL, MB_EX_ILLEGAL_ADDRESS);
    return;
  }

  //Check for failure ?? Why ?
  /*
  if (coil (reg) != (bool) status) {
    exceptionResponse (MB_FC_WRITE_COIL, MB_EX_SLAVE_FAILURE);
    return;
  }
  */

  _reply = MB_REPLY_ECHO; // reply with received frame
}

/* Func 15: Write Multiple Coils
    Request
      Function code       1 Byte  0x0F
      Starting Address    2 Bytes 0x0000 to 0xFFFF
      Quantity of Outputs 2 Bytes 0x0001 to 0x07B0
      Byte Count          1 Byte  N* (Quantity of Outputs / 8, if the remainder is different of 0 -> N = N+1)
      Outputs Value       N* x 1 Byte
    Response
      Function code       1 Byte  0x0F
      Starting Address    2 Bytes 0x0000 to 0xFFFF
      Quantity of Outputs 2 Bytes 0x0001 to 0x07B0
*/
//-------------------------------------------------------------------------------
// private
void Modbus::writeMultipleCoils (byte *frame, word startreg, word numoutputs, byte bytecount) {
  //Check value
  word bytecount_calc = numoutputs / 8;
  if (numoutputs % 8) {
    bytecount_calc++;
  }
  if (numoutputs < 0x0001 || numoutputs > 0x07B0 || bytecount != bytecount_calc) {
    exceptionResponse (MB_FC_WRITE_COILS, MB_EX_ILLEGAL_VALUE);
    return;
  }

  //Check Address (startreg...startreg + numregs)
  for (word k = 0; k < numoutputs; k++) {
    if (!searchRegister (startreg + TRegister::CoilOffset + k)) {
      exceptionResponse (MB_FC_WRITE_COILS, MB_EX_ILLEGAL_ADDRESS);
      return;
    }
  }

  //Clean frame buffer
  free (_frame);
  _len = 5;
  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_WRITE_COILS, MB_EX_SLAVE_FAILURE);
    return;
  }

  _frame[0] = MB_FC_WRITE_COILS;
  _frame[1] = startreg >> 8;
  _frame[2] = startreg & 0x00FF;
  _frame[3] = numoutputs >> 8;
  _frame[4] = numoutputs & 0x00FF;

  byte bitn = 0;
  word totoutputs = numoutputs;
  word i;
  while (numoutputs--) {
    i = (totoutputs - numoutputs) / 8;
    setCoil (startreg, bitRead (frame[6 + i], bitn));
    //increment the bit index
    bitn++;
    if (bitn == 8) {
      bitn = 0;
    }
    //increment the register
    startreg++;
  }

  _reply = MB_REPLY_NORMAL;
}

/*
  Func 17:  Report Server ID
    Request
      Function code         1 Byte            0x11
    Response
      Function code         1 Byte            0x11
      Byte Count            1 Byte
      Server ID             device specific
      Run Indicator Status  1 Byte            0x00 = OFF, 0xFF = ON
      Additional Data
*/
//-------------------------------------------------------------------------------
// private
void Modbus::reportServerId() {
  //Clean frame buffer
  free (_frame);
  _len = 4;
  if (_additional_data) {
    _len += strlen (_additional_data);
  }
  _frame = (byte *) malloc (_len);
  if (!_frame) {
    exceptionResponse (MB_FC_REPORT_SERVER_ID, MB_EX_SLAVE_FAILURE);
    return;
  }
  _frame[0] = MB_FC_REPORT_SERVER_ID;
  _frame[1] = _len - 2;   //byte count
  _frame[2] = 0x00; // Server ID
  _frame[3] = 0xFF; // Run Indicator Status
  if (_additional_data) { // Additional Data
    strncpy ( (char *) &_frame[4], _additional_data, strlen (_additional_data));
  }
  _reply = MB_REPLY_NORMAL;
}

//-------------------------------------------------------------------------------
int Modbus::setAdditionalServerData (const char data[]) {

  free (_additional_data);
  if (data) {
    size_t l = strlen (data) + 1;

    _additional_data = (char *) malloc (l);
    if (_additional_data) {

      strcpy (_additional_data, data);
      return l;
    }
  }
  return 0;
}

//-------------------------------------------------------------------------------
#endif // USE_HOLDING_REGISTERS_ONLY not defined
